# This file was auto-generated by update_ml_package_versions.py.
# Please do not edit it manually.

_ML_PACKAGE_VERSIONS = {
    "sklearn": {
        "package_info": {
            "pip_release": "scikit-learn"
        },
        "models": {
            "minimum": "0.22.1",
            "maximum": "1.2.2"
        },
        "autologging": {
            "minimum": "0.22.1",
            "maximum": "1.2.2"
        }
    },
    "pytorch": {
        "package_info": {
            "pip_release": "torch"
        },
        "models": {
            "minimum": "1.6.0",
            "maximum": "2.0.0"
        },
        "autologging": {
            "minimum": "1.6.0",
            "maximum": "2.0.0"
        }
    },
    "pytorch-lightning": {
        "package_info": {
            "pip_release": "pytorch-lightning"
        },
        "autologging": {
            "minimum": "1.0.5",
            "maximum": "2.0.1.post0"
        }
    },
    "tensorflow": {
        "package_info": {
            "pip_release": "tensorflow"
        },
        "models": {
            "minimum": "2.3.0",
            "maximum": "2.12.0"
        },
        "autologging": {
            "minimum": "2.3.0",
            "maximum": "2.12.0"
        }
    },
    "xgboost": {
        "package_info": {
            "pip_release": "xgboost"
        },
        "models": {
            "minimum": "1.1.1",
            "maximum": "1.7.5"
        },
        "autologging": {
            "minimum": "1.1.1",
            "maximum": "1.7.5"
        }
    },
    "lightgbm": {
        "package_info": {
            "pip_release": "lightgbm"
        },
        "models": {
            "minimum": "2.3.1",
            "maximum": "3.3.5"
        },
        "autologging": {
            "minimum": "2.3.1",
            "maximum": "3.3.5"
        }
    },
    "catboost": {
        "package_info": {
            "pip_release": "catboost"
        },
        "models": {
            "minimum": "0.23.1",
            "maximum": "1.1.1"
        }
    },
    "gluon": {
        "package_info": {
            "pip_release": "mxnet"
        },
        "models": {
            "minimum": "1.5.1",
            "maximum": "1.9.1"
        },
        "autologging": {
            "minimum": "1.5.1",
            "maximum": "1.9.1"
        }
    },
    "fastai": {
        "package_info": {
            "pip_release": "fastai"
        },
        "models": {
            "minimum": "2.4.1",
            "maximum": "2.7.12"
        },
        "autologging": {
            "minimum": "2.4.1",
            "maximum": "2.7.12"
        }
    },
    "onnx": {
        "package_info": {
            "pip_release": "onnx"
        },
        "models": {
            "minimum": "1.7.0",
            "maximum": "1.13.1"
        }
    },
    "spacy": {
        "package_info": {
            "pip_release": "spacy"
        },
        "models": {
            "minimum": "2.2.4",
            "maximum": "3.5.2"
        }
    },
    "statsmodels": {
        "package_info": {
            "pip_release": "statsmodels"
        },
        "models": {
            "minimum": "0.11.1",
            "maximum": "0.13.5"
        },
        "autologging": {
            "minimum": "0.11.1",
            "maximum": "0.13.5"
        }
    },
    "spark": {
        "package_info": {
            "pip_release": "pyspark"
        },
        "models": {
            "minimum": "3.0.0",
            "maximum": "3.4.0"
        },
        "autologging": {
            "minimum": "3.0.0",
            "maximum": "3.4.0"
        }
    },
    "mleap": {
        "package_info": {
            "pip_release": "mleap"
        },
        "models": {
            "minimum": "0.18.0",
            "maximum": "0.22.0"
        }
    },
    "prophet": {
        "package_info": {
            "pip_release": "prophet"
        },
        "models": {
            "minimum": "1.0.1",
            "maximum": "1.1.2"
        }
    },
    "pmdarima": {
        "package_info": {
            "pip_release": "pmdarima"
        },
        "models": {
            "minimum": "1.8.0",
            "maximum": "2.0.3"
        }
    },
    "diviner": {
        "package_info": {
            "pip_release": "diviner"
        },
        "models": {
            "minimum": "0.1.0",
            "maximum": "0.1.1"
        }
    },
    "h2o": {
        "package_info": {
            "pip_release": "h2o"
        },
        "models": {
            "minimum": "3.40.0.1",
            "maximum": "3.40.0.3"
        }
    },
    "shap": {
        "package_info": {
            "pip_release": "shap"
        },
        "models": {
            "minimum": "0.41.0",
            "maximum": "0.41.0"
        }
    },
    "paddle": {
        "package_info": {
            "pip_release": "paddlepaddle"
        },
        "models": {
            "minimum": "2.4.1",
            "maximum": "2.4.2"
        }
    },
    "transformers": {
        "package_info": {
            "pip_release": "transformers"
        },
        "models": {
            "minimum": "4.25.1",
            "maximum": "4.28.1"
        }
    },
    "openai": {
        "package_info": {
            "pip_release": "openai"
        },
        "models": {
            "minimum": "0.27.2",
            "maximum": "0.27.4"
        }
    },
    "langchain": {
        "package_info": {
            "pip_release": "langchain"
        },
        "models": {
            "minimum": "0.0.139",
            "maximum": "0.0.143"
        }
    }
}
